import random
from importlib import resources
from typing import List, Optional, Dict

import yaml

from v2share.base import BaseConfig
from v2share.data import V2Data
from v2share.exceptions import TransportNotSupportedError, ProtocolNotSupportedError

supported_transports = ["tcp", "http", "ws", "grpc", "h2"]
supported_protocols = ["vmess", "trojan", "shadowsocks"]


class ClashConfig(BaseConfig):
    def __init__(self, template_path: Optional[str] = None, swallow_errors=True):
        if not template_path:
            template_path = resources.files("v2share.templates") / "clash.yml"
        with open(template_path) as f:
            self.template_data = f.read()
        self._swallow_errors = swallow_errors
        self._configs = []

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in supported_transports
            ) or (unsupported_protocol := proxy.protocol not in supported_protocols):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)

    def render(self, sort: bool = True, shuffle: bool = False):
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda config: config.weight)
        else:
            configs = self._configs

        proxies, remarks = [], []
        for proxy in configs:
            proxies.append(self._get_node(proxy))
            remarks.append(proxy.remark)

        result = yaml.safe_load(self.template_data)
        result["proxies"] = proxies
        result["rules"] = []
        result["proxy-groups"][0]["proxies"] = remarks
        return yaml.safe_dump(result, sort_keys=False)

    def _make_node(
        self,
        name: str,
        protocol: str,
        server: str,
        port: int,
        network: str,
        tls: bool,
        sni: str,
        host: str,
        path: str,
        udp: bool = True,
        alpn: str = "",
        ais: bool = "",
    ):
        if protocol == "shadowsocks":
            protocol = "ss"

        node = {
            "name": name,
            "type": protocol,
            "server": server,
            "port": port,
            "network": network,
            f"{network}-opts": {},
            "udp": udp,
        }

        if type == "ss":  # shadowsocks
            return node

        if tls:
            node["tls"] = True
            if type == "trojan":
                node["sni"] = sni
            else:
                node["servername"] = sni
            if alpn:
                node["alpn"] = alpn.split(",")
            if ais:
                node["skip-cert-verify"] = ais

        net_opts = node[f"{network}-opts"]

        if network == "ws":
            if path:
                net_opts["path"] = path
            if host:
                net_opts["headers"] = {"Host": host}

        if network == "grpc":
            if path:
                net_opts["grpc-service-name"] = path

        if network == "h2":
            if path:
                net_opts["path"] = path
            if host:
                net_opts["host"] = [host]

        if network in {"http", "tcp"}:
            if path:
                net_opts["method"] = "GET"
                net_opts["path"] = [path]
            if host:
                net_opts["method"] = "GET"
                net_opts["headers"] = {"Host": host}

        return node

    def _get_node(self, config: V2Data) -> Dict:
        node = self._make_node(
            name=config.remark,
            protocol=config.protocol,
            server=config.address,
            port=config.port,
            network=config.transport_type,
            tls=(config.tls == "tls"),
            sni=config.sni,
            host=config.host,
            path=config.path,
            udp=True,
            alpn=config.alpn,
            ais=config.allow_insecure,
        )

        if config.protocol == "vmess":
            node["uuid"] = str(config.uuid)
            node["alterId"] = 0
            node["cipher"] = "auto"
        elif config.protocol == "trojan":
            node["password"] = config.password
        elif config.protocol == "shadowsocks":
            node["password"] = config.password
            node["cipher"] = config.shadowsocks_method

        return node



import base64
import json
import urllib.parse as urlparse
from dataclasses import dataclass, field
from typing import Optional, Dict
from uuid import UUID

from v2share._utils import filter_dict
from v2share.exceptions import ProtocolNotSupportedError


@dataclass
class V2Data:
    protocol: str
    remark: str
    address: str
    port: int
    shadowsocks_method: Optional[str] = None
    uuid: Optional[UUID] = None
    vmess_security: str = "auto"
    password: Optional[str] = None
    ed25519: Optional[str] = None
    host: Optional[str] = None
    http_headers: Dict[str, str] = field(default_factory=dict)
    transport_type: str = "tcp"
    grpc_multi_mode: bool = False
    path: Optional[str] = None
    header_type: Optional[str] = None
    tls: str = "none"
    flow: Optional[str] = None
    sni: Optional[str] = None
    fingerprint: Optional[str] = None
    alpn: Optional[str] = None
    reality_pbk: Optional[str] = None
    reality_sid: Optional[str] = None
    reality_spx: Optional[str] = None
    client_address: Optional[str] = None
    fragment: bool = False
    fragment_packets: str = "tlshello"
    fragment_length: str = "100-200"
    fragment_interval: str = "10-20"
    enable_mux: bool = False
    allow_insecure: bool = False
    weight: int = 1

    def _apply_tls_settings(self, payload):
        if self.tls in ["tls", "reality"]:
            payload.update(
                {
                    "sni": self.sni,
                    "fp": self.fingerprint,
                }
            )
            if self.protocol in ["vless", "trojan"]:
                payload.update({"allowInsecure": int(self.allow_insecure) or None})
        if self.tls == "tls":
            payload.update(
                {
                    "alpn": self.alpn,
                }
            )
        elif self.tls == "reality":
            payload.update(
                {
                    "pbk": self.reality_pbk,
                    "sid": self.reality_sid,
                    "spx": self.reality_spx,
                }
            )

    def _apply_trojan_vless_transport(self, payload):
        if self.transport_type == "grpc":
            transport_data = {
                "serviceName": self.path,
                "authority": self.host,
                "mode": "multi" if self.grpc_multi_mode else "gun",
            }
        elif self.transport_type == "kcp":
            transport_data = {"seed": self.path}
        elif self.transport_type == "quic":
            transport_data = {"key": self.path, "quicSecurity": self.host}
        else:
            transport_data = {"path": self.path, "host": self.host}

        payload.update(transport_data)

    def to_link(self):
        if self.protocol == "shadowsocks":
            return (
                "ss://"
                + base64.b64encode(
                    (
                        f"{self.shadowsocks_method}:{self.password}"
                        + f"@{self.address}:{self.port}"
                    ).encode()
                ).decode()
                + f"#{urlparse.quote(self.remark)}"
            )

        if self.protocol == "vmess":
            payload = {
                "add": self.address,
                "aid": "0",
                "host": self.host,
                "id": str(self.uuid),
                "net": self.transport_type,
                "path": self.path,
                "port": self.port,
                "ps": self.remark,
                "scy": self.vmess_security,
                "tls": self.tls,
                "type": self.header_type,
                "v": "2",
            }

            self._apply_tls_settings(payload)
            payload = filter_dict(payload, ("", None))
            return (
                "vmess://"
                + base64.b64encode(
                    json.dumps(payload, sort_keys=True).encode()
                ).decode()
            )

        if self.protocol in ["trojan", "vless"]:
            payload = {
                "security": self.tls,
                "type": self.transport_type,
                "headerType": self.header_type,
            }
            if self.protocol == "vless" and self.flow:
                payload.update({"flow": self.flow or ""})

            self._apply_trojan_vless_transport(payload)

            self._apply_tls_settings(payload)

            payload = filter_dict(payload, ("", None))

            passphrase = (
                self.uuid
                if self.protocol == "vless"
                else urlparse.quote(self.password, safe=":")
            )
            return (
                f"{self.protocol}://"
                + f"{passphrase}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )

        if self.protocol == "hysteria2":
            payload = {"sni": self.sni}
            if self.header_type:
                payload.update({"obfs": self.header_type})
                payload.update({"obfs-password": self.path})
            if self.allow_insecure:
                payload.update({"insecure": "1"})

            return (
                "hysteria2://"
                + f"{self.password}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )

        if self.protocol == "wireguard":
            payload = {"publickey": self.path, "address": self.client_address}
            return (
                    "wireguard://"
                    + f"{self.ed25519}@{self.address}:{self.port}?"
                    + urlparse.urlencode(payload)
                    + f"#{(urlparse.quote(self.remark))}"
            )
        raise ProtocolNotSupportedError



"""v2share utilities"""

from typing import Dict, Sequence, Any


def filter_dict(d: Dict[Any, Any], values: Sequence[Any]) -> dict:
    """
    removes pairs with null value from dictionary
    :param d: the dictionary
    :param values: values you want to be filtered
    :return: the filtered dictionary
    """
    return dict(filter(lambda p: p[1] not in values, d.items()))



from .clash import ClashConfig
from .clashmeta import ClashMetaConfig
from .data import V2Data
from .singbox import SingBoxConfig
from .xray import XrayConfig

__all__ = ["V2Data", "SingBoxConfig", "ClashConfig", "ClashMetaConfig", "XrayConfig"]



import json
import random
from importlib import resources
from typing import List

from v2share.base import BaseConfig
from v2share.data import V2Data
from v2share.exceptions import ProtocolNotSupportedError, TransportNotSupportedError

supported_transports = [
    "tcp",
    "kcp",
    "mkcp",
    "ws",
    "websocket",
    "http",
    "h2",
    "quic",
    "grpc",
    "gun",
    "httpupgrade",
    "splithttp",
]
supported_protocols = ["vmess", "vless", "trojan", "shadowsocks"]


class XrayConfig(BaseConfig):
    def __init__(
        self,
        template_path: str = None,
        mux_template_path: str = None,
        swallow_errors=True,
    ):
        self.config = []
        self._configs = []
        self._swallow_errors = swallow_errors
        if not template_path:
            template_path = resources.files("v2share.templates") / "xray.json"
        if not mux_template_path:
            mux_template_path = resources.files("v2share.templates") / "xray_mux.json"
        with open(template_path) as f:
            self._template = f.read()
        with open(mux_template_path) as f:
            self._mux_template = f.read()

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in supported_transports
            ) or (unsupported_protocol := proxy.protocol not in supported_protocols):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)

    def render(self, sort: bool = True, shuffle: bool = False):
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda config: config.weight)
        else:
            configs = self._configs

        xray_configs = []
        for data in configs:
            outbound = {"tag": data.remark, "protocol": data.protocol}

            if data.protocol == "vmess":
                outbound["settings"] = XrayConfig.vmess_config(
                    address=data.address, port=data.port, uuid=str(data.uuid)
                )
            elif data.protocol == "vless":
                if data.tls in {"reality", "tls"}:
                    flow = data.flow or ""
                else:
                    flow = ""
                outbound["settings"] = XrayConfig.vless_config(
                    address=data.address, port=data.port, uuid=str(data.uuid), flow=flow
                )

            elif data.protocol == "trojan":
                outbound["settings"] = XrayConfig.trojan_config(
                    address=data.address, port=data.port, password=data.password
                )

            elif data.protocol == "shadowsocks":
                outbound["settings"] = XrayConfig.shadowsocks_config(
                    address=data.address,
                    port=data.port,
                    password=data.password,
                    method=data.shadowsocks_method,
                )

            outbounds = [outbound]
            dialer_proxy = None

            if data.fragment:
                fragment_outbound = XrayConfig.make_fragment_outbound(
                    data.fragment_packets, data.fragment_length, data.fragment_interval
                )
                outbounds.append(fragment_outbound)
                dialer_proxy = fragment_outbound["tag"]

            outbound["streamSettings"] = XrayConfig.make_stream_settings(
                net=data.transport_type,
                tls=data.tls,
                sni=data.sni,
                host=data.host,
                path=data.path,
                alpn=data.alpn,
                fp=data.fingerprint,
                pbk=data.reality_pbk,
                sid=data.reality_sid,
                ais=data.allow_insecure,
                header_type=data.header_type,
                grpc_multi_mode=data.grpc_multi_mode,
                dialer_proxy=dialer_proxy,
                headers=data.http_headers,
            )

            mux_config = json.loads(self._mux_template)
            mux_config["enabled"] = data.enable_mux
            outbound["mux"] = mux_config

            json_template = json.loads(self._template)
            complete_config = {
                **json_template,
                **{
                    "remarks": data.remark,
                    "outbounds": outbounds + json_template["outbounds"],
                },
            }
            xray_configs.append(complete_config)
        return json.dumps(xray_configs, indent=4)

    @staticmethod
    def tls_config(sni, fingerprint, alpn=None, ais=False):
        tlsSettings = {"serverName": sni, "allowInsecure": ais or False}
        if alpn is None:
            alpn = ["h2", "http/1.1"]

        tlsSettings["alpn"] = alpn

        if fingerprint:
            tlsSettings["fingerprint"] = fingerprint

        return tlsSettings

    @staticmethod
    def reality_config(public_key, short_id, sni, fingerprint="", spiderx=""):

        return {
            "serverName": sni,
            "fingerprint": fingerprint,
            "show": False,
            "publicKey": public_key,
            "shortId": short_id,
            "spiderX": spiderx,
        }

    @staticmethod
    def ws_config(path=None, host=None, headers=None):
        if headers is None:
            headers = {}

        ws_settings = {"headers": headers}
        if path:
            ws_settings["path"] = path
        if host:
            ws_settings["host"] = host

        return ws_settings

    @staticmethod
    def httpupgrade_config(path=None, host=None, headers=None):
        if headers is None:
            headers = {}

        httpupgrade_settings = {"headers": headers}
        if path:
            httpupgrade_settings["path"] = path
        if host:
            httpupgrade_settings["host"] = host

        return httpupgrade_settings

    @staticmethod
    def grpc_config(authority=None, service_name=None, multi_mode=False):

        grpc_settings = {
            "multiMode": multi_mode,
        }
        if service_name:
            grpc_settings["serviceName"] = service_name
        if authority:
            grpc_settings["authority"] = authority
        return grpc_settings

    @staticmethod
    def tcp_http_config(header_type=None):
        if header_type is None:
            header_type = "none"

        tcp_settings = {
            "header": {
                "type": header_type,
            }
        }

        return tcp_settings

    @staticmethod
    def h2_config(path=None, host=None, headers=None):
        if host is None:
            host = []
        if path is None:
            path = "/"
        if headers is None:
            headers = {}

        http_settings = {"path": path, "host": host}
        if headers:
            http_settings["headers"] = headers

        return http_settings

    @staticmethod
    def splithttp_config(path=None, host=None, headers=None):
        if host is None:
            host = []
        if path is None:
            path = "/"
        if headers is None:
            headers = {}

        splithttp_settings = {"path": path, "headers": headers}

        if host:
            splithttp_settings["host"] = host

        return splithttp_settings

    @staticmethod
    def quic_config(security="none", key=None, header_type="none"):

        quic_settings = {
            "security": security,
            "header": {"type": header_type},
        }

        if key:
            quic_settings["key"] = key

        return quic_settings

    @staticmethod
    def kcp_config(seed=None, header_type=None):

        kcp_settings = {
            "header": {},
        }

        if seed:
            kcp_settings["seed"] = seed
        if header_type:
            kcp_settings["header"]["type"] = header_type

        return kcp_settings

    @staticmethod
    def make_stream_settings(
        net="tcp",
        path=None,
        host=None,
        tls="none",
        sni="",
        fp=None,
        alpn="h2,http/1.1",
        pbk=None,
        sid=None,
        ais=False,
        header_type=None,
        grpc_multi_mode=False,
        dialer_proxy=None,
        headers=None,
    ):
        if headers is None:
            headers = {}

        stream_settings = {"network": net}

        if net in {"ws", "websocket"}:
            stream_settings["wsSettings"] = XrayConfig.ws_config(
                path=path, host=host, headers=headers
            )
        elif net in {"grpc", "gun"}:
            stream_settings["grpcSettings"] = XrayConfig.grpc_config(
                authority=host, service_name=path, multi_mode=grpc_multi_mode
            )
        elif net in {"h2", "http"}:
            stream_settings["httpSettings"] = XrayConfig.h2_config(
                path=path, host=[host], headers=headers
            )
        elif net in {"kcp", "mkcp"}:
            stream_settings["kcpSettings"] = XrayConfig.kcp_config(
                seed=path, header_type=header_type
            )
        elif net == "tcp":
            stream_settings["tcpSettings"] = XrayConfig.tcp_http_config(
                header_type=header_type
            )
        elif net == "quic":
            stream_settings["quicSettings"] = XrayConfig.quic_config(
                security=host, key=path, header_type=header_type
            )
        elif net == "httpupgrade":
            stream_settings["httpupgradeSettings"] = XrayConfig.httpupgrade_config(
                path=path, host=host, headers=headers
            )
        elif net == "splithttp":
            stream_settings["splithttpSettings"] = XrayConfig.splithttp_config(
                path=path, host=host, headers=headers
            )

        if tls == "tls":
            stream_settings["security"] = "tls"
            stream_settings["tlsSettings"] = XrayConfig.tls_config(
                sni=sni, fingerprint=fp, alpn=alpn.split(",") if alpn else None, ais=ais
            )
        elif tls == "reality":
            stream_settings["security"] = "reality"
            stream_settings["realitySettings"] = XrayConfig.reality_config(
                pbk, sid, sni, fingerprint=fp
            )
        else:
            stream_settings["security"] = "none"

        if dialer_proxy:
            stream_settings["sockopt"] = {"dialerProxy": dialer_proxy}

        return stream_settings

    @staticmethod
    def vmess_config(address: str, port: int, uuid: str):

        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "security": "auto",
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def vless_config(address, port, uuid, flow=""):

        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "encryption": "none",
                            "flow": flow,
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def trojan_config(address, port, password):
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                }
            ]
        }

    @staticmethod
    def shadowsocks_config(address, port, password, method):
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                    "method": method,
                    "uot": False,
                }
            ]
        }

    @staticmethod
    def make_fragment_outbound(packets, length, interval):
        return {
            "tag": "fragment_out",
            "protocol": "freedom",
            "settings": {
                "fragment": {"packets": packets, "length": length, "interval": interval}
            },
        }



from abc import ABC, abstractmethod
from typing import List

from .data import V2Data


class BaseConfig(ABC):
    @abstractmethod
    def render(self, sort: bool, shuffle: bool) -> str:
        pass

    @abstractmethod
    def add_proxies(self, proxies: List[V2Data]) -> None:
        pass



import random
from typing import List

from v2share.base import BaseConfig
from v2share.data import V2Data
from v2share.exceptions import TransportNotSupportedError, ProtocolNotSupportedError

supported_transports = [
    "tcp",
    "kcp",
    "ws",
    "http",
    "quic",
    "grpc",
    "httpupgrade",
    "splithttp",
    None,
]
supported_protocols = ["vmess", "vless", "trojan", "shadowsocks", "hysteria2", "wireguard"]


class LinksConfig(BaseConfig):
    def __init__(self, swallow_errors=True):
        self._configs: List[V2Data] = []
        self._swallow_errors = swallow_errors

    def render(self, sort: bool = True, shuffle: bool = False):
        if shuffle is True:
            configs = random.sample(self._configs, len(self._configs))
        elif sort is True:
            configs = sorted(self._configs, key=lambda config: config.weight)
        else:
            configs = self._configs

        links = [config.to_link() for config in configs]
        return "\n".join(links)

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            # validation
            if (
                unsupported_transport := proxy.transport_type
                not in supported_transports
            ) or (unsupported_protocol := proxy.protocol not in supported_protocols):
                if self._swallow_errors:
                    continue
                if unsupported_transport:
                    raise TransportNotSupportedError
                if unsupported_protocol:
                    raise ProtocolNotSupportedError

            self._configs.append(proxy)
